/********************************************
*
* Copyright (c) 2022-2024 AO MCST
* Distributed under the terms of MIT License.
*
*********************************************/

#ifndef _MGA2_DRM_H_
#define _MGA2_DRM_H_ 1

#include <stdint.h>
#include <stdbool.h>

#include <stdio.h>

/* DEBUG options: */
//#define MGA2_DRM_DEBUG  /* DRM debug trace */
//#define MGA2_DRM_DUMP_AUC2_PACKETS


#define MGA2_BO_VRAM    0x00000001
#define MGA2_BO_CPU     0x00000002
#define MGA2_BO_RD      0x00000100
#define MGA2_BO_WR      0x00000200
#define MGA2_BO_RDWR    (MGA2_BO_RD | MGA2_BO_WR)

/* Chip definitions */
#define PCI_VENDOR_MCST     0x1fff
#define PCI_CHIP_MGA2       0x8009
#define PCI_CHIP_MGA25      0x8031
#define PCI_CHIP_MGA26      0x803a

/* http://wiki.lab.sun.mcst.ru/e2kwiki/P2/Memory_subsystem
  размер видеопамяти, выделяемой из системной памяти
  может быть размером 128M, 256M, 512M и 1024M, однако
  будет использовать некоторое минимальное значение в
  случае, если у ядра не удалось выяснить реальный размер*/

#define MGA2_VRAM_SIZE_DEFAULT    70*1024*1024ul /*   70 MiB */
#define MGA25_VRAM_SIZE_DEFAULT 1024*1024*1024ul /* 1024 MiB */
#define MGA26_VRAM_SIZE_DEFAULT  512*1024*1024ul /*  512 MiB */


#ifdef MGA2_DRM_DEBUG
extern int printOffset;

#define MGA2_DRM_DBG_ENTER()                                    \
        do {                                                    \
            fprintf(stderr, "[libdrm] ");                                   \
            fprintf(stderr, "%*c %s\n", ++printOffset, '>', __func__);   \
        } while (0)

#define MGA2_DRM_DBG_PRINT(fmt, ...)    \
        do {                            \
            fprintf(stderr, "[libdrm] ");           \
            fprintf(stderr, fmt, __VA_ARGS__);   \
        } while (0)

#define MGA2_DRM_DBG_LEAVE(...)                                 \
        do {                                                    \
            fprintf(stderr, "[libdrm] ");                                   \
            fprintf(stderr, "%*c %s\n", printOffset--, '<', __func__);   \
            return __VA_ARGS__;                                 \
        } while (0)

#define ERROR_MSG(fmt, ...) \
        do { fprintf(stderr, "[EE] [libdrm] " fmt " (%s:%d)\n", \
                ##__VA_ARGS__, __FUNCTION__, __LINE__); } while (0)
#else
#define MGA2_DRM_DBG_ENTER()
#define MGA2_DRM_DBG_PRINT(fmt, ...)
#define MGA2_DRM_DBG_LEAVE(...) return __VA_ARGS__
#define ERROR_MSG(fmt, ...)
#endif

// --------------------------------------------------------------------------

struct mga2_list {
    struct mga2_list *prev;
    struct mga2_list *next;
};

// --------------------------------------------------------------------------

struct mga2_object {
    struct mga2_object *parent;
    uint64_t handle;
};

struct mga2_drm {
    struct mga2_object client;
    int fd;
    uint32_t version;
    uint64_t vram_size;
    unsigned int rings_num; /* Number of BCTRL/AUC2 rings */
};

static inline struct mga2_drm *
mga2_drm(struct mga2_object *obj)
{
    while (obj && obj->parent)
        obj = obj->parent;
    return (struct mga2_drm *)obj;
}

int mga2_drm_new(int fd, struct mga2_drm **pdrm);
void mga2_drm_del(struct mga2_drm **pdrm);

// --------------------------------------------------------------------------

struct mga2_device {
    struct mga2_object object;
    uint64_t vram_limit;
    uint64_t vram_used;
    uint32_t drm_version;
    int fd;
};

struct mga2_device_priv {
    struct mga2_device base;
    int close;
    pthread_mutex_t lock;
    struct mga2_list bo_list;
    int nr_client;
    uint32_t *client;
};

static inline struct mga2_device_priv *
mga2_device(struct mga2_device *dev)
{
    return (struct mga2_device_priv *)dev;
}

int mga2_device_new(struct mga2_object *parent,
                      struct mga2_device **pdev);

int mga2_device_wrap(int fd, int close, struct mga2_device **pdev);
int mga2_device_open(const char *busid, struct mga2_device **pdev);
void mga2_device_del(struct mga2_device **pdev);

// --------------------------------------------------------------------------

struct mga2_client {
    struct mga2_device *device;
    int id;
};

struct mga2_client_priv {
    struct mga2_client base;
    /* TODO: ?*/
};

static inline struct mga2_client_priv *
mga2_client(struct mga2_client *client)
{
    return (struct mga2_client_priv *) client;
}

int mga2_client_new(struct mga2_device *dev, struct mga2_client **pclient);
void mga2_client_del(struct mga2_client **pclient);

// --------------------------------------------------------------------------

struct mga2_bo_config {
    uint32_t memtype;
    uint32_t data[8];
};

struct mga2_bo {
    struct mga2_device *device;
    uint32_t handle;
    uint64_t size;
    uint32_t flags;
    uint64_t offset;
    void *map;
    uint32_t memtype;
};

int mga2_bo_new(struct mga2_device *dev, uint32_t flags, uint32_t align,
                  uint64_t size, struct mga2_bo_config *config,
                  struct mga2_bo **pbo);

int mga2_bo_wrap(struct mga2_device *dev, uint32_t handle,
                   struct mga2_bo **pbo);

void mga2_bo_ref(struct mga2_bo *bo, struct mga2_bo **pref);

int mga2_bo_map(struct mga2_bo *bo, uint32_t access,
                  struct mga2_client *client);

int mga2_bo_wait(struct mga2_bo *bo, uint32_t access,
                   struct mga2_client *client);

int mga2_bo_name_get(struct mga2_bo *bo, char *name);

// --------------------------------------------------------------------------

#define DRM_IOCTL_MGA2_BCTRL        DRM_IOW(DRM_COMMAND_BASE  + 0, struct mga2_drm_bctrl)
#define DRM_IOCTL_MGA2_GEM_NEW      DRM_IOWR(DRM_COMMAND_BASE + 1, struct mga2_drm_gem_new)
#define DRM_IOCTL_MGA2_GEM_MMAP     DRM_IOWR(DRM_COMMAND_BASE + 2, struct mga2_drm_mga2_gem_mmap)
#define DRM_IOCTL_MGA2_SYNC         DRM_IO(DRM_COMMAND_BASE   + 3)
#define DRM_IOCTL_MGA2_INFO         DRM_IOWR(DRM_COMMAND_BASE  + 4, struct drm_mga2_info)
#define DRM_IOCTL_MGA2_AUC2         DRM_IOW(DRM_COMMAND_BASE  + 5, struct mga2_drm_bctrl) /* Add decriptor to AUC2 ring */

#define MGA2_GEM_DOMAIN_CPU         0x1
#define MGA2_GEM_DOMAIN_VRAM        0x2

struct mga2_drm_bctrl {
    uint32_t handle;
    uint32_t pad;
    uint64_t buffers;
};

struct mga2_drm_gem_new {
    uint64_t size;
    uint32_t domain;
    uint32_t handle;
};

struct mga2_drm_mga2_gem_mmap {
    uint32_t handle;
    uint32_t pad;
    uint64_t offset;
};

// --------------------------------------------------------------------------

/**
 * Bits for control package
 */
#define CTRL_PACKAGE_LAST_CMD_MASK      (1<<15)
#define CTRL_PACKAGE_WAIT_BLT_DONE      (1<<1)
#define CTRL_PACKAGE_WAIT_BLT_FIFO      (1<<0)

#define MGA2_DRM_RING_COMMAND_BUFFER_SIZE   8192
#define MGA2_DRM_RING_WORDS                 (MGA2_DRM_RING_COMMAND_BUFFER_SIZE / 4)

#ifndef ARRAY_SIZE
    #define ARRAY_SIZE(x)   (sizeof(x)/sizeof(x[0]))
#endif

struct mga2_drm_buffers {
    uint16_t nr;
    uint16_t reltype; /* 0 - low half of address, 1 - high*/
    uint32_t handle;
    uint32_t offset[0];
};

struct mga2_ring_buffer {
    struct mga2_bo *buffer;     /* Pointer to command words buffer */
    uint32_t handle;            /* Handle to the pointer to command words buffer */
    uint32_t current;           /* Current position in the pointer to command words buffer */
    uint32_t expected_end;      /* Expected end of packet */
    uint32_t last_header;       /* Position of last header */

    /* Table with relocation information */
    struct {
        struct mga2_drm_buffers header;
        uint32_t                offset[1];
    } reloc_info[MGA2_DRM_RING_WORDS];

    uint32_t ridx;      /* Index of the first free record in relocation table */
    int fd;             /* DRM handle */
};

/* Input structure for the INFO ioctl */
struct drm_mga2_info {
        /* Where the return value will be stored */
        __u64 return_pointer;
        /* The size of the return value. Just like "size" in "snprintf",
         * it limits how many bytes the kernel can write. */
        __u32 return_size;
        /* The query request id. */
        __u32 query;

        union {
                struct {
                        __u32 _reserved[4];
                } _reserved;
        };
};

/* Query memory about VRAM and GTT domains */
#define MGA2_INFO_MEMORY                        0x1

struct drm_mga2_heap_info {
        /** max. physical memory */
        __u64 total_heap_size;

        /** Theoretical max. available memory in the given heap */
        __u64 usable_heap_size;

        /**
         * Number of bytes allocated in the heap. This includes all processes
         * and private allocations in the kernel. It changes when new buffers
         * are allocated, freed, and moved. It cannot be larger than
         * heap_size.
         */
        __u64 heap_usage;

        /**
         * Theoretical possible max. size of buffer which
         * could be allocated in the given heap
         */
        __u64 max_allocation;
};

struct drm_mga2_memory_info {
        struct drm_mga2_heap_info vram;
        struct drm_mga2_heap_info cpu_accessible_vram;
        struct drm_mga2_heap_info gtt;
};

#define AUC2_NONLASTCMD_MASK (1<<31)

struct mga2_rings {
    struct mga2_ring_buffer *ring[256 + 0];
    struct mga2_ring_buffer *current_ring;
    unsigned int rings_num;
    unsigned int idx;
    bool flush;
};

int mga2_init_rings(struct mga2_device *dev, struct mga2_client *client, struct mga2_rings **prings);
void mga2_release_rings(struct mga2_rings **prings);

void FLUSH_RING(struct mga2_rings *rings);
void BEGIN_RING(struct mga2_rings *rings, int size);
void END_RING(struct mga2_rings *rings);
void RELOC_RING(struct mga2_ring_buffer *buf, uint32_t handle);
void PUSH_RING(struct mga2_ring_buffer *buf, uint32_t data);


void FLUSH_AUC2_TASKSLIST(struct mga2_rings *rings);
void BEGIN_AUC2_PACKET(struct mga2_rings *rings, int size);
void PUSH_AUC2_WORD32(struct mga2_ring_buffer *buf, uint32_t data);
void FINISH_AUC2_PACKET(struct mga2_rings *rings);
void RELOC_AUC2_PACKET(struct mga2_ring_buffer *buf, uint32_t handle, int reltype);

#endif /* _MGA2_DRM_H_ */
